/**
 * USART driver.
 * @author Ryan Hall
 */
component USART {

    // bit names for UCSRnA register
    field RXCn:  8 = 0x80;  // USART receive complete
    field TXCn:  8 = 0x40;  // USART transmit complete
    field UDREn: 8 = 0x20;  // USART data registe empty
    field FEn:   8 = 0x10;  // Frame error : set to 1 when writing to UCSRnA
    field DORn:  8 = 0x08;  // Data over run : set to 0 when writing to UCSRnA
    field UPEn:  8 = 0x04;  // Parity error : set to 0 when writing to UCSRnA
    field U2Xn:  8 = 0x02;  // Double USART transmission speed : set to 0 in sync mode
    field MPCMn: 8 = 0x01;  // Multi-processor communication mode

    // bit names for UCSRnB register
    field RXCIEn: 8 = 0x80;  // enable receive complete interrupt
    field TXCIEn: 8 = 0x40;  // enable transmission complete interrupt
    field UDRIEn: 8 = 0x20;  // enable data register empty interrupt
    field RXENn:  8 = 0x10;  // enable receiver
    field TXENn:  8 = 0x08;  // enable transmitter
    field UCSZn2: 8 = 0x04;  // high end bit of 3 for frame size
    field RXB8n:  8 = 0x02;  // 9th bit of received character
    field TXB8n:  8 = 0x01;  // 9th bit of transmitted character

    // bit names for UCSRnC register
    field UMSELn: 8 = 0x40;  // 0 for async, 1 for sync mode
    field UPMn1:  8 = 0x20;  // Parity mode
    field UPMn0:  8 = 0x10;  // Parity mode
    field USBSn:  8 = 0x08;  // Number of stop bits inserted by transmitter
    field UCSZn1: 8 = 0x04;  // Character size low 2 bits
    field UCSZn0: 8 = 0x02;  // Character size low 2 bits
    field UCPOLn: 8 = 0x01;  // Clock polarity in sync mode

    // Baud Rate Registers
    // UBRRnL : lower 8 bits of baud rate
    // UBRRnH : higher 4 bits of baud rate

    // Data Register
    // UDRn

    method calculateBaud(fosc: int, rate: int, mode: int): int {
	if ( mode == 0 ) return fosc / (16 * rate) - 1;
	if ( mode == 1 ) return fosc / (8 * rate) - 1;
	if ( mode == 2 ) return fosc / (2 * rate) - 1;
	return 0;
    }
}

component USART0 {

    //bit names for UCSR0A register
    field RXC0: 8 = 0x80;  //USART receive complete
    field TXC0: 8 = 0x40;  //USART transmit complete
    field UDRE0: 8 = 0x20;  //USART data registe empty
    field FE0: 8 = 0x10;  //Frame error : set to 1 when writing to UCSR0A
    field DOR0: 8 = 0x08;  //Data over run : set to 0 when writing to UCSR0A
    field UPE0: 8 = 0x04;  //Parity error : set to 0 when writing to UCSR0A
    field U2X0: 8 = 0x02;  //Double USART transmission speed : set to 0 in sync mode
    field MPCM0: 8 = 0x01;  //Multi-processor communication mode

    //bit names for UCSR0B register
    field RXCIE0: 8 = 0x80;  //enable receive complete interrupt
    field TXCIE0: 8 = 0x40;  //enable transmission complete interrupt
    field UDRIE0: 8 = 0x20;  //enable data register empty interrupt
    field RXEN0: 8 = 0x10;  //enable receiver
    field TXEN0: 8 = 0x08;  //enable transmitter
    field UCSZ02: 8 = 0x04; //high end bit of 3 for frame size
    field RXB80: 8 = 0x02;  //9th bit of received character
    field TXB80: 8 = 0x01;  //9th bit of transmitted character

    //bit names for UCSR0C register
    field UMSEL0: 8 = 0x40;  //0 for async, 1 for sync mode
    field UPM01: 8 = 0x20;  //Parity mode
    field UPM00: 8 = 0x10;  //Parity mode
    field USBS0: 8 = 0x08;  //Number of stop bits inserted by transmitter
    field UCSZ01: 8 = 0x04;  //Charcter size low 2 bits
    field UCSZ00: 8 = 0x02;  //Charcter size low 2 bits
    field UCPOL0: 8 = 0x01;  //Clock polarity in sync mode

    field r_queue : Queue;  //receive queue
    field t_queue : Queue;  //transmit queue
     
    constructor() {
        r_queue = new Queue(10);
        t_queue = new Queue(10);
    }

    method enable(baud_rate: int, sync_mode : boolean, double_async : boolean) {

        // set character size to 8
        device.UCSR0C = UCSZ01 | UCSZ00;
     
        if (sync_mode) {
            device.UCSR0C = device.UCSR0C | UMSEL0;
        } else if (double_async) {
            device.UCSR0A = U2X0;
        }
      
        // set baud rate
        device.UBRR0H = (baud_rate >> 8) :: 8;
        device.UBRR0L = (baud_rate & 0xFF) :: 8;
 
        // enable interrupts, receiver and transmitter
        device.UCSR0B = RXEN0 | TXEN0 | RXCIE0;
    }

    method disable() {
       // disable transmitter and receiver
       device.UCSR0B = 0x0;
    }

    method transmit(buf : char[], i : int, l : int) : int {
        local index : int = 0;
	local mask = MCU.disable();
        for (index = 0 ; index < l ; index++) {
            if (t_queue.enqueue(buf[i + index])) { 
                if (t_queue.size() > 0) {
                    // turn on data ready interrupts
                    device.UCSR0B = device.UCSR0B | UDRIE0; 
                }
            } else {
		MCU.restore(mask);
                return index;
            }
        }
	MCU.restore(mask);
        return index;    
    }

    method tran_handler() {
        local c = t_queue.dequeue();
        if (t_queue.size() == 0) {
            // turn off data ready interrupts
            device.UCSR0B = device.UCSR0B & ~UDRIE0;
        }
        device.UDR0 = c;
    }

    method receive(buf : char[], i : int, l : int) : int {
        local count : int = 0;
	local mask = MCU.disable();
        while (count < l and r_queue.size() > 0) {
            buf[i + count] = r_queue.dequeue();        
            count++;
        }
	MCU.restore(mask);
        return count;
    }

    method rec_handler() {
        local c : char = (device.UDR0) :: char;
        r_queue.enqueue(c);
    }
}

component USART1 {

    //bit names for UCSR0A register
    field RXC1: 8 = 0x80;  //USART receive complete
    field TXC1: 8 = 0x40;  //USART transmit complete
    field UDRE1: 8 = 0x20;  //USART data registe empty
    field FE1: 8 = 0x10;  //Frame error : set to 1 when writing to UCSR0A
    field DOR1: 8 = 0x08;  //Data over run : set to 0 when writing to UCSR0A
    field UPE1: 8 = 0x04;  //Parity error : set to 0 when writing to UCSR0A
    field U2X1: 8 = 0x02;  //Double USART transmission speed : set to 0 in sync mode
    field MPCM1: 8 = 0x01;  //Multi-processor communication mode

    //bit names for UCSR0B register
    field RXCIE1: 8 = 0x80;  //enable receive complete interrupt
    field TXCIE1: 8 = 0x40;  //enable transmission complete interrupt
    field UDRIE1: 8 = 0x20;  //enable data register empty interrupt
    field RXEN1: 8 = 0x10;  //enable receiver
    field TXEN1: 8 = 0x08;  //enable transmitter
    field UCSZ12: 8 = 0x04; //high end bit of 3 for frame size
    field RXB81: 8 = 0x02;  //9th bit of received character
    field TXB81: 8 = 0x01;  //9th bit of transmitted character

    //bit names for UCSR0C register
    field UMSEL1: 8 = 0x40;  //0 for async, 1 for sync mode
    field UPM11: 8 = 0x20;  //Parity mode
    field UPM10: 8 = 0x10;  //Parity mode
    field USBS1: 8 = 0x08;  //Number of stop bits inserted by transmitter
    field UCSZ11: 8 = 0x04;  //Charcter size low 2 bits
    field UCSZ10: 8 = 0x02;  //Charcter size low 2 bits
    field UCPOL1: 8 = 0x01;  //Clock polarity in sync mode

    field r_queue : Queue;  //receive queue
    field t_queue : Queue;  //transmit queue
     
    constructor() {
        r_queue = new Queue(10);
        t_queue = new Queue(10);
    }

    method enable(baud_rate: int, sync_mode : boolean, double_async : boolean) {

        // set character size to 8
        device.UCSR1C = UCSZ11 | UCSZ10;
     
        if (sync_mode) {
            device.UCSR1C = device.UCSR1C | UMSEL1;
        } else if(double_async) {
            device.UCSR1A = U2X1;
        }
      
        // set baud rate
        device.UBRR1H = (baud_rate >> 8) :: 8;
        device.UBRR1L = (baud_rate & 0xFF) :: 8;
 
        // enable interrupts, receiver and transmitter
        device.UCSR1B = RXEN1 | TXEN1 | RXCIE1;
    }

    method disable() {
        // disable receiver, transmitter and reciever
        device.UCSR1B = 0x0;
    }

    method transmit(buf : char[], i : int, l : int) : int {
        local index : int = 0;
	local mask = MCU.disable();
        for (index = 0 ; index < l ; index++) {
            if (t_queue.enqueue(buf[i + index])) { 
                if (t_queue.size() > 0) {
                    // turn on data ready interrupts
                   device.UCSR1B = device.UCSR1B | UDRIE1; 
                }
            } else {
		MCU.restore(mask);
                return index;
            }
        }
	MCU.restore(mask);
        return index;    
    }

    method tran_handler() {
        local c = t_queue.dequeue();
        if (t_queue.size() == 0) {
            //turn off data ready interrupts
            device.UCSR1B = device.UCSR1B & ~UDRIE1;
        }
        device.UDR1 = c;
    }

    method receive(buf : char[], i : int, l : int) : int {
        local count : int = 0;
	local mask = MCU.disable();
        while(count < l and r_queue.size() > 0) {
            buf[i + count] = r_queue.dequeue();        
            count++;
        }
	MCU.restore(mask);
        return count;
    }

    method rec_handler() {
        local c : char = (device.UDR1) :: char;
        r_queue.enqueue(c);
    }
}
